;**********************************************************
;         TEA: The Tiny Encryption Algorithm
;                PIC16 implementation
;
;Written by: Marco DI LEO <m.dileo@sistinf.it>
;Date: 1998/04/21
;Version: 1.1
;
;The following code is Copyright (C) 1998 by Marco DI LEO.
;Permission is hereby granted to use the code for
;non-commercial personal use providing this notice is not
;removed from source code copies. Contact me if you wish to
;use the code (or part of it) for a commercial use. 
;This source code is provided AS-IS with no warranty
;expressed or implied.
;**********************************************************

;History:
;V1.0 1998/04/17 First release. Works! User can choose how many
;rounds to perform changing the 'roundin' equate. All affected
;values are calculated accordingly. Uses all indirect variable
;access (FSR/INDF) for the math operations. A bit slow.
;V1.1 1998/04/21 Greatly improved speed. Saved some RAM register.
;Uses macro instead of function for math. Created a separate
;'math32.inc' file that can be reused for other programs.

;Based on The Tiny Encryption Algorithm (TEA) by David Wheeler
;and Roger Needham of the Cambridge Computer Laboratory.

;TEA is a Feistel cipher with XOR and addition as the non-linear
;mixing functions. 
;Takes 64 bits of data in 'vy' and 'vz' and 128 bits of key in 'ka',
;'kb', 'kc' and 'kd'. 'roundin' is the number of iterations. 32 is
;ample, 16 is sufficient, as few as eight may be OK. The algorithm
;achieves good dispersion after only six iterations.
;'delta' is chosen to be the real part of (the golden ratio
;Sqrt(5/4) - 1/2 ~ 0.618034 multiplied by 2^32).

;To use the routines include this file in the main program. Be sure to
;declare a 'cblock' (even an empty one) with proper memory address before
;the 'include' line. Set the plain/ciphertext registers with the data to
;be encrypted or decrypted, put the key in ka...kd and call TEAenc or TEAdec.
;The resulting cipher/plaintext will be in the vy, vz register.

	include	"math32.inc"		;32 bit math routines

roundin	equ	D'32'			;Number of rounds for enc/decipher
delta	equ	0x9E3779B9		;Magic value used by TEA
sumin	equ	(delta * roundin)	;Used in the decipher to initialize 'sum'

	cblock				;Incremental memory allocation
		vy: 4, vz: 4		;Plain/ciphertext block
		ka: 4, kb: 4		;Key block
		kc: 4, kd: 4
		sum: 4			;Internal variable for TEA
		round			;Internal round counter
		tmp2: 4			;Internal temp register
		tmp1: 4			;Internal temp register
	endc


;Encipher a block of plaintext

;Takes the plaintext in 'vy' and 'vz', the 128 bit key in 'ka', 'kb', 'kc'
;and 'kd'. After execution the ciphertext will be in 'vy' and 'vz'. The
;value of the key registers is unchanged.

TEAenc	mmovl32	sum, 0		;sum = 0

	movlw	roundin		;Preload the round counter
	movwf	round

loope	maddl32	sum, delta, sum	;Compute sum += delta

	call	TEAy		;Compute the right part of vy += ..... in tmp1

	madd32	vy, tmp1, vy	;vy += (vz << 4) + ka ^ vz + sum ^ (vz >> 5) + kb

	call	TEAz		;Compute the right part of vz += ..... in tmp1

	madd32	vz, tmp1, vz	;vz += (vy << 4) + kc ^ vy + sum ^ (vy >> 5) + kd

	decfsz	round, F	;round--
	goto	loope

	return


;Decipher a block of ciphertext

;Takes the ciphertext in 'vy' and 'vz', the 128 bit key in 'ka', 'kb', 'kc'
;and 'kd'. After execution the plaintext will be in 'vy' and 'vz'. The
;value of the key registers is unchanged.

TEAdec	mmovl32	sum, sumin 	;sum = sumin

	movlw	roundin		;Preload the round counter
	movwf	round

loopd	call	TEAz		;Compute the right part of vz -= ..... in tmp1

	msub32	vz, tmp1, vz	;vz -= (vy << 4) + kc ^ vy + sum ^ (vy >> 5) + kd

	call	TEAy		;Compute the right part of vy -= ..... in tmp1

	msub32	vy, tmp1, vy	;vy -= (vz << 4) + ka ^ vz + sum ^ (vz >> 5) + kb

	msubl32	sum, delta, sum	;sum -= delta

	decfsz	round, F	;round--
	goto	loopd

	return


;Computes the value ((vz << 4) + ka) ^ (vz + sum) ^ ((vz >> 5) + kb)

;The value is used in both the encipher and decipher routines

TEAy	swapf	vz, W		;Performs tmp1 = vz << 4 and tmp2 = vz >> 4
	andlw	0xf0		;using the swapf nibble swap instruction.
	movwf	tmp1		;tmp1 LSB...

	swapf	vz, W		;...tmp1 LSB + 1 and tmp2 LSB...
	andlw	0x0f
	movwf	tmp1+1
	movwf	tmp2
	swapf	vz+1, W
	andlw	0xf0
	iorwf	tmp1+1, F
	iorwf	tmp2, F

	swapf	vz+1, W		;...tmp1 LSB + 2 and tmp2 LSB + 1...
	andlw	0x0f
	movwf	tmp1+2
	movwf	tmp2+1
	swapf	vz+2, W
	andlw	0xf0
	iorwf	tmp1+2, F
	iorwf	tmp2+1, F

	swapf	vz+2, W		;...tmp1 LSB + 3 (MSB) and tmp2 LSB + 2...
	andlw	0x0f
	movwf	tmp1+3
	movwf	tmp2+2
	swapf	vz+3, W
	andlw	0xf0
	iorwf	tmp1+3, F
	iorwf	tmp2+2, F

	swapf	vz+3, W		;...tmp2 LSB + 3 (MSB)
	andlw	0x0f
	movwf	tmp2+3

	bcf	STATUS, C	;Do a one bit shift right of tmp2 to get
	rrf	tmp2+3, F	;the required tmp2 = vz >> 5
	rrf	tmp2+2, F
	rrf	tmp2+1, F
	rrf	tmp2, F

	madd32	tmp1, ka, tmp1	;tmp1 = (vz << 4) + ka

	madd32	tmp2, kb, tmp2	;tmp2 = (vz >> 5) + kb

	call	xor32		;tmp1 = (vz << 4) + ka ^ (vz >> 5) + kb

	madd32	vz, sum, tmp2	;tmp2 = vz + sum

	call	xor32		;tmp1 = (vz << 4) + ka ^ vz + sum ^ (vz >> 5) + kb

	return


;Computes the value ((vy << 4) + kc) ^ (vy + sum) ^ ((vy >> 5) + kd)

;The value is used in both the encipher and decipher routines

TEAz	swapf	vy, W		;Performs tmp1 = vy << 4 and tmp2 = vy >> 4
	andlw	0xf0		;using the swapf nibble swap instruction.
	movwf	tmp1		;tmp1 LSB...

	swapf	vy, W		;...tmp1 LSB + 1 and tmp2 LSB...
	andlw	0x0f
	movwf	tmp1+1
	movwf	tmp2
	swapf	vy+1, W
	andlw	0xf0
	iorwf	tmp1+1, F
	iorwf	tmp2, F

	swapf	vy+1, W		;...tmp1 LSB + 2 and tmp2 LSB + 1...
	andlw	0x0f
	movwf	tmp1+2
	movwf	tmp2+1
	swapf	vy+2, W
	andlw	0xf0
	iorwf	tmp1+2, F
	iorwf	tmp2+1, F

	swapf	vy+2, W		;...tmp1 LSB + 3 (MSB) and tmp2 LSB + 2...
	andlw	0x0f
	movwf	tmp1+3
	movwf	tmp2+2
	swapf	vy+3, W
	andlw	0xf0
	iorwf	tmp1+3, F
	iorwf	tmp2+2, F

	swapf	vy+3, W		;...tmp2 LSB + 3 (MSB)
	andlw	0x0f
	movwf	tmp2+3

	bcf	STATUS, C	;Do a one bit shift right of tmp2 to get
	rrf	tmp2+3, F	;the required tmp2 = vz >> 5
	rrf	tmp2+2, F
	rrf	tmp2+1, F
	rrf	tmp2, F

	madd32	tmp1, kc, tmp1	;tmp1 = (vy << 4) + kc

	madd32	tmp2, kd, tmp2	;tmp2 = (vy >> 5) + kd

	call	xor32		;tmp1 = (vy << 4) + kc ^ (vy >> 5) + kd

	madd32	vy, sum, tmp2	;tmp2 = vy + sum

	call	xor32		;tmp1 = (vy << 4) + kc ^ vy + sum ^ (vy >> 5) + kd

	return



;Performs the exclusive-or of two 32 bit values.

;Computes the ex-or of the value in the tmp2 register and tmp1
;register. The result is stored back in tmp1.

xor32	movf	tmp2, W		;Get operand LSB
	xorwf	tmp1, F		;Xor with working register LSB
	movf	tmp2+1, W	;Get operand LSB + 1
	xorwf	tmp1+1, F
	movf	tmp2+2, W	;Get operand LSB + 2
	xorwf	tmp1+2, F
	movf	tmp2+3, W	;Get operand LSB + 3 (MSB)
	xorwf	tmp1+3, F
	return

